/* Copyright (C) 2002-2003 RealVNC Ltd.  All Rights Reserved.
 *    
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */
#include <stdio.h>
#include <vector>
#include <rdr/types.h>
#include <rfb/Exception.h>
#include <rfb/ComparingUpdateTracker.h>

using namespace rfb;

ComparingUpdateTracker::ComparingUpdateTracker(const PixelBuffer* buffer)
  : SimpleUpdateTracker(true), fb(buffer),
    oldFb(fb->getPF(), fb->width(), fb->height())
{
  int bytesPerPixel = fb->getPF().bpp/8;
  int newStrideBytes = fb->getStride() * bytesPerPixel;
  int oldStrideBytes = oldFb.getStride() * bytesPerPixel;

  for (int i = 0; i < fb->height(); i++) {
    memcpy((rdr::U8*)oldFb.data + i * oldStrideBytes,
           (rdr::U8*)fb->data + i * newStrideBytes,
           oldStrideBytes);
  }
}

ComparingUpdateTracker::~ComparingUpdateTracker()
{
}

void ComparingUpdateTracker::flush_update(UpdateInfo &info,
                                          const Region &cliprgn)
{
  throw rfb::Exception("flush_update(UpdateInfo&) not implemented");
}

void ComparingUpdateTracker::flush_update(UpdateTracker &ut,
                                          const Region &cliprgn)
{
  throw rfb::Exception("flush_update(UpdateTracker&) not implemented");
}

void ComparingUpdateTracker::compare()
{
  std::vector<Rect> rects;
  std::vector<Rect>::iterator i;

  copied.get_rects(&rects, copy_delta.x<=0, copy_delta.y<=0);
  for (i = rects.begin(); i != rects.end(); i++)
    oldFb.copyRect(*i, copy_delta);

  Region to_check = changed.union_(copied);
  to_check.get_rects(&rects);

  Region newChanged;
  for (i = rects.begin(); i != rects.end(); i++)
    compareRect(*i, &newChanged);

  copied.assign_subtract(newChanged);
  changed = newChanged;
}

#define BLOCK_SIZE 16

void ComparingUpdateTracker::compareRect(const Rect& r, Region* newChanged)
{
  if (!r.enclosed_by(fb->getRect())) {
    fprintf(stderr,"ComparingUpdateTracker: rect outside fb (%d,%d-%d,%d)\n", r.tl.x, r.tl.y, r.br.x, r.br.y);
    return;
  }

  int bytesPerPixel = fb->getPF().bpp/8;
  int newStrideBytes = fb->getStride() * bytesPerPixel;
  int oldStrideBytes = oldFb.getStride() * bytesPerPixel;

  std::vector<Rect> changedBlocks;

  for (int blockTop = r.tl.y; blockTop < r.br.y; blockTop += BLOCK_SIZE)
  {
    rdr::U8* newBlockPtr = (((rdr::U8*)fb->data) + blockTop * newStrideBytes
                            + r.tl.x * bytesPerPixel);
    rdr::U8* oldBlockPtr = (((rdr::U8*)oldFb.data) + blockTop * oldStrideBytes
                            + r.tl.x * bytesPerPixel);

    int blockBottom = std::min(blockTop+BLOCK_SIZE, r.br.y);
    for (int blockLeft = r.tl.x; blockLeft < r.br.x; blockLeft += BLOCK_SIZE)
    {
      rdr::U8* newPtr = newBlockPtr;
      rdr::U8* oldPtr = oldBlockPtr;

      int blockRight = std::min(blockLeft+BLOCK_SIZE, r.br.x);
      int blockWidthInBytes = (blockRight-blockLeft) * bytesPerPixel;

      for (int y = blockTop; y < blockBottom; y++)
      {
        if (memcmp(oldPtr, newPtr, blockWidthInBytes) != 0)
        {
          // A block has changed - copy the remainder to the oldFb
          changedBlocks.push_back(Rect(blockLeft, blockTop,
                                       blockRight, blockBottom));
          for (int y2 = y; y2 < blockBottom; y2++)
          {
            memcpy(oldPtr, newPtr, blockWidthInBytes);
            newPtr += newStrideBytes;
            oldPtr += oldStrideBytes;
          }
          break;
        }

        newPtr += newStrideBytes;
        oldPtr += oldStrideBytes;
      }

      oldBlockPtr += blockWidthInBytes;
      newBlockPtr += blockWidthInBytes;
    }
  }

  if (!changedBlocks.empty()) {
    Region temp;
    temp.setOrderedRects(changedBlocks);
    newChanged->assign_union(temp);
  }
}
