/* Copyright (C) 2002-2003 RealVNC Ltd.  All Rights Reserved.
 *    
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

// -=- PixelBuffer.cxx
//
// The PixelBuffer class encapsulates the PixelFormat and dimensions
// of a block of pixel data.

#include <rfb/Exception.h>
#include <rfb/LogWriter.h>
#include <rfb/PixelBuffer.h>

using namespace rfb;
using namespace rdr;

static LogWriter vlog("PixelBuffer");


// -=- Generic pixel buffer class

PixelBuffer::PixelBuffer(const PixelFormat& pf, int w, int h, rdr::U8* data_,
                         ColourMap* cm)
  : data(data_), format(pf), width_(w), height_(h), colourmap(cm) {}
PixelBuffer::PixelBuffer() : data(0), width_(0), height_(0), colourmap(0) {}

PixelBuffer::~PixelBuffer() {}

void PixelBuffer::setPF(const PixelFormat &pf) {format = pf;}
const PixelFormat& PixelBuffer::getPF() const {return format;}

int PixelBuffer::getStride() const {return width();}

ColourMap* PixelBuffer::getColourMap() const {return colourmap;}

void
PixelBuffer::fillRect(const Rect& r, Pixel pix) {
  int bytesPerPixel = getPF().bpp/8;
  int bytesPerRow = bytesPerPixel * getStride();
  int bytesPerFill = bytesPerPixel * r.width();
  U8* pos = data + (r.tl.y * bytesPerRow) + (r.tl.x * bytesPerPixel);
  for (int i=r.tl.y; i<r.br.y; i++) {
    switch (bytesPerPixel) {
    case 1:
      memset(pos, pix, bytesPerFill);
      break;
    case 2:
      {
        U16* opix = (U16*) pos;
        U16* eol = opix + r.width();
        while (opix < eol)
          *opix++ = pix;
      }
      break;
    case 4:
      {
        U32* opix = (U32*) pos;
        U32* eol = opix + r.width();
        while (opix < eol)
          *opix++ = pix;
      }
      break;
    default:
      vlog.debug("unsupported pixel size:%u", bytesPerPixel);
      throw rdr::Exception("unsupported pixel size");
    }
    pos += bytesPerRow;
  }
}

void
PixelBuffer::imageRect(const Rect& r, void* pixels) {
  int bytesPerPixel = getPF().bpp/8;
  int bytesPerDestRow = bytesPerPixel * getStride();
  int bytesPerSrcRow = bytesPerPixel * r.width();
  U8* dest_pos = data + (r.tl.y * bytesPerDestRow) + (r.tl.x * bytesPerPixel);
  U8* src_pos = (U8*)pixels;
  for (int i=r.tl.y; i<r.br.y; i++) {
    memcpy(dest_pos, src_pos, bytesPerSrcRow);
    dest_pos += bytesPerDestRow;
    src_pos += bytesPerSrcRow;
  }
}

void
PixelBuffer::maskRect(const Rect& r, void* pixels, void* mask_) {
  char* mask = (char*) mask_;
  int w = r.width();
  int h = r.height();
  int maskBytesPerRow = (w + 7) / 8;

  int stride = getStride();

  for (int y = 0; y < h; y++) {
    int cy = r.tl.y + y;
    if (cy >= 0 && cy < height()) {
      for (int x = 0; x < w; x++) {
        int cx = r.tl.x + x;
        if (cx >= 0 && cx < width()) {
          int byte = y * maskBytesPerRow + x / 8;
          int bit = 7 - x % 8;
          if (mask[byte] & (1 << bit)) {
            switch (getPF().bpp) {
            case 8:
              ((U8*)data)[cy * stride + cx] = ((U8*)pixels)[y * w + x];
              break;
            case 16:
              ((U16*)data)[cy * stride + cx] = ((U16*)pixels)[y * w + x];
              break;
            case 32:
              ((U32*)data)[cy * stride + cx] = ((U32*)pixels)[y * w + x];
              break;
            }
          }
        }
      }
    }
  }
}

void
PixelBuffer::copyRect(const Rect &rect, const Point &move_by_delta) {
  // We assume that the specified rectangle is pre-clipped to the buffer
  unsigned int bytesPerPixel, bytesPerRow, bytesPerMemCpy;
  Rect srect = rect.translate(move_by_delta.negate());
  bytesPerPixel = format.bpp/8;
  bytesPerRow = getStride() * bytesPerPixel;
  bytesPerMemCpy = rect.width() * bytesPerPixel;
  if (move_by_delta.y <= 0) {
    U8* dest = data + rect.tl.x*bytesPerPixel + rect.tl.y*bytesPerRow;
    U8* src = data + srect.tl.x*bytesPerPixel + srect.tl.y*bytesPerRow;
    for (int i=rect.tl.y; i<rect.br.y; i++) {
      memmove(dest, src, bytesPerMemCpy);
      dest += bytesPerRow;
      src += bytesPerRow;
    }
  } else {
    U8* dest = data + rect.tl.x*bytesPerPixel + (rect.br.y-1)*bytesPerRow;
    U8* src = data + srect.tl.x*bytesPerPixel + (srect.br.y-1)*bytesPerRow;
    for (int i=rect.tl.y; i<rect.br.y; i++) {
      memmove(dest, src, bytesPerMemCpy);
      dest -= bytesPerRow;
      src -= bytesPerRow;
    }
  }
}

void
PixelBuffer::getImage(void* imageBuf, const Rect& r, int outStride) const {
  // We assume that the specified rectangle is pre-clipped to the buffer
  int bytesPerPixel = format.bpp/8;
  int inBytesPerRow = getStride() * bytesPerPixel;
  if (!outStride) outStride = r.width();
  int outBytesPerRow = outStride * bytesPerPixel;
  int bytesPerMemCpy = r.width() * bytesPerPixel;
  U8* imageBufPos = (U8*)imageBuf;
  U8* dataPos = data + r.tl.x*bytesPerPixel + r.tl.y*inBytesPerRow;
  unsigned int h=r.height();
  for (unsigned int i=0; i<h;i++) {
    memcpy(imageBufPos, dataPos, bytesPerMemCpy);
    imageBufPos += outBytesPerRow;
    dataPos += inBytesPerRow;
  }
}

Pixel PixelBuffer::getPixel(const Point& p) {
  switch(format.bpp) {
  case 8: return *((rdr::U8*)getDataAt(p));
  case 16: return *((rdr::U16*)getDataAt(p));
  case 32: return *((rdr::U32*)getDataAt(p));
  default: return 0;
  };
}


// -=- Managed pixel buffer class
// Automatically allocates enough space for the specified format & area

ManagedPixelBuffer::ManagedPixelBuffer()
  : datasize(0), own_colourmap(false)
{
  checkDataSize();
};

ManagedPixelBuffer::ManagedPixelBuffer(const PixelFormat& pf, int w, int h)
  : PixelBuffer(pf, w, h, 0, 0), datasize(0), own_colourmap(false) {
  checkDataSize();
};

ManagedPixelBuffer::~ManagedPixelBuffer() {
  if (data) delete [] data;
  if (colourmap && own_colourmap) delete colourmap;
};

void
ManagedPixelBuffer::setPF(const PixelFormat &pf) {
  format = pf; checkDataSize();
};
void
ManagedPixelBuffer::setSize(int w, int h) {
  width_ = w; height_ = h; checkDataSize();
};

void
ManagedPixelBuffer::setColourMap(ColourMap* cm, bool own_cm) {
  if (colourmap && own_colourmap) delete colourmap;
  colourmap = cm;
  own_colourmap = own_cm;
}

inline void
ManagedPixelBuffer::checkDataSize() {
  unsigned long new_datasize = width_ * height_ * (format.bpp/8);
  if (datasize < new_datasize) {
    vlog.debug("reallocating managed buffer");
    if (data) {
      delete [] data;
      datasize = 0; data = 0;
    }
    if (new_datasize) {
      data = new U8[new_datasize];
      if (!data)
        throw Exception("rfb::ManagedPixelBuffer unable to allocate buffer");
      datasize = new_datasize;
    }
  }
};
