/* Copyright (C) 2002-2003 RealVNC Ltd.  All Rights Reserved.
 *    
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

// -=- PixelBuffer.h
//
// The PixelBuffer class encapsulates the PixelFormat and dimensions
// of a block of pixel data.

#ifndef __RFB_PIXEL_BUFFER_H__
#define __RFB_PIXEL_BUFFER_H__

#include <rfb/ImageGetter.h>
#include <rfb/PixelFormat.h>
#include <rfb/ColourMap.h>
#include <rfb/Rect.h>
#include <rfb/Pixel.h>

namespace rfb {

  class Region;

  class PixelBuffer : public ImageGetter {
  public:
    PixelBuffer(const PixelFormat& pf, int width, int height, rdr::U8* data_,
                ColourMap* cm);
    virtual ~PixelBuffer();

    // - Set the pixel buffer layout
    virtual void setPF(const PixelFormat &pf);
    virtual const PixelFormat &getPF() const;

    int width()  const { return width_; }
    int height() const { return height_; }
    int area() const { return width_ * height_; }

    // getRect() returns a rectangle the same size as this PixelBuffer whose
    // top-left is at the given point, or at (0,0).
    Rect getRect() const { return Rect(0, 0, width_, height_); }
    Rect getRect(const Point& pos) const {
      return Rect(pos, pos.translate(Point(width_, height_)));
    }

    // - Get the number of pixels per row in the actual pixel buffer data area
    //   This may in some cases NOT be the same as width().
    virtual int getStride() const;

    // grabRegion() causes the pixel data in the given region to be "grabbed".
    // For a PixelBuffer not associated with any kind of device, this is a null
    // operation.
    virtual void grabRegion(const Region& region) {}

    // - Return the ColourMap associated with the buffer, if palettised.
    virtual ColourMap* getColourMap() const;

    // - Some simple drawing operations
    //   Fill a rectangle, put pixel data in a rectangle, copy a rectangle &
    //   put pixel data in a rectangle using a mask
    virtual void fillRect(const Rect &dest, Pixel pix);
    virtual void imageRect(const Rect &dest, void* pixels);
    virtual void copyRect(const Rect &dest, const Point &move_by_delta);
    virtual void maskRect(const Rect& r, void* pixels, void* mask_);

    // - Copy a rectangle of pixel data into the supplied buffer
    virtual void getImage(void* imageBuf, const Rect& r, int stride=0) const;

    // - Get a pointer to the pixel data starting at (x,y).
    rdr::U8* getDataAt(const Point& p) {
      return data + (getStride() * p.y + p.x) * format.bpp/8;
    }

    // - Get the data at (x,y) as a Pixel
    Pixel getPixel(const Point& p);

    rdr::U8* data;
  protected:
    PixelBuffer();
    PixelFormat format;
    int width_, height_;
    ColourMap* colourmap;
  };

  // -=- Managed pixel buffer class
  // Automatically allocates enough space for the specified format & area

  class ManagedPixelBuffer : public PixelBuffer {
  public:
    ManagedPixelBuffer();
    ManagedPixelBuffer(const PixelFormat& pf, int width, int height);
    virtual ~ManagedPixelBuffer();

    // - Changed the managed buffer layout - ensures the data area is large
    //   enough
    virtual void setPF(const PixelFormat &pf);
    virtual void setSize(int w, int h);

    // - Assign a ColourMap to the managed buffer.
    //   If own_cm is true then the ColourMap is owned by the managed buffer.
    virtual void setColourMap(ColourMap* cm, bool own_cm);

    // dataLen() returns the length in bytes of the pixel data
    int dataLen() const { return width_ * height_ * (format.bpp/8); }

  protected:
    unsigned long datasize;
    bool own_colourmap;
    void checkDataSize();
  };

};

#endif // __RFB_PIXEL_BUFFER_H__
